#!/usr/bin/env python

'''
store calibration data for forces
dictionary with key =	"normal": 4 by a+1 array for normal force,
						"shear"	: 2 by b+1 array for shear force
a = order of normal fit, which is 3
b = order of shear fit, which is 2
'''

import json
import numpy as np

'''
calibration names DO NOT include .json
eg. "COM2" : "original"
'''
sensor_calibrations = {
	"original" : "original", # special case for testing
	"/dev/tty.usbserial-A6WMSVHL" : "original",
	"/dev/tty.usbserial-A8WERMKN" : "original",
	"/dev/tty.usbserial-A6WMNFNP" : "original",
	"/dev/tty.usbserial-A5VAF3WT" : "original", # demo board
	"COM22" : "original",
	"COM24" : "original",
	"COM26" : "original",
	"COM4" : "sensor2_4", # Sensor 2 at JPL
        "COM10" : "sensor2_me112", # Sensor 2 at JPL
        "/dev/ttyUSB0" : "sensor2_me112",
}

normal_num = 4
normal_fit_order = 3
shear_num = 2
shear_fit_order = 2

def EmptyCalibration(name):
	sensor = {}

	sensor["name"] = name

	normal_arr = [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]] #[0] * normal_num
	sensor["normal"] = normal_arr

	shear_arr = [[0, 0, 0], [0, 0, 0]] #[0] * shear_num
	sensor["shear"] = shear_arr

	return sensor

def LoadCalibration(dev_name):

	cal_name = sensor_calibrations[dev_name]
	filename = cal_name + ".json"

	with open(filename, "r") as f:
		cal = json.load(f)

	if cal["name"] != cal_name:
		raise Exception("Calibration name does not match")

	return cal

def SaveCalibration(name, cal=None):
	if not cal:
		cal = EmptyCalibration(name)

	filename = name + ".json"

	json_cal = json.dumps(cal, sort_keys=True, indent=4, separators=(',', ': '))
	with open(filename, "w") as f:
		f.write(json_cal)

''' x is a scalar and fit is a vector of coefficients '''
def polyfit(x, fit):
	sum = 0
	for i in range(len(fit)):
		sum += np.power(x,i) * fit[i]
	#sum = x*fit[1] + x^2*fit[2] + x^3*fit[3]
	return sum

'''
assumes indata is in the format:
[Sx1, Sx2, Sy1, Sy2, T1, T2, T3, T4]

outputs in the format:
[Sx, Sy, T1, T2, T3, T4]
'''
def CalculateForce(indata, calibration):

	data = list(indata)

	if len(data) != 8:
		raise IndexError("data only has {0} of 8 elements".format(len(data)))

	shear = [data[1] - data[0], data[3] - data[2]]
	normal = data[4:]

	for i in range(len(normal)):
		#normal[i] = polyfit(normal[i], calibration["normal"][i])
		normal[i] = polyfit(normal[i], calibration["normal"][i])

	for i in range(len(shear)):
		shear[i] = polyfit(shear[i], calibration["shear"][i])

	return shear + normal


#orginal = {
#	"name" : "original",
#	"normal" :	[[-0.00553099, 0.000339965, 2.74158e-08, 8.11113e-13],
#				 [-0.00553099, 0.000339965, 2.74158e-08, 8.11113e-13],
#				 [-0.00553099, 0.000339965, 2.74158e-08, 8.11113e-13],
#				 [-0.00553099, 0.000339965, 2.74158e-08, 8.11113e-13]],
#	"shear" :	[[-0.057098, -0.0062031],
#				 [-0.00058174, -0.00012549]]
#}
#
#
#
#if __name__ == '__main__':
#
#	# SaveCalibration("empty")
#	# SaveCalibration("original", orginal)
#
#	cal = LoadCalibration("COM22")
#
#	# print cal
#
#	print cal["normal"]

