#! /usr/bin/python

import matplotlib
matplotlib.use('TkAgg')

import argparse
import copy
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np
import sensor
import serial
import serial.tools.list_ports
import signal
import sys
import threading

# Read flags
parser = argparse.ArgumentParser(description="Save and/or plot live data coming over a serial connection")
parser.add_argument("-b", "--baudrate", default=115200, type=int, help="Set the baudrate of the serial connection")
parser.add_argument("-d", "--disable-axes", default=False, const=True, action="store_const", help="Disable axis labeling to speed up plotting")
parser.add_argument("-k", "--keep", default=2, type=float, help="Set the number of seconds of data to keep")
parser.add_argument("-l", "--live-plot", default=False, const=True, action="store_const", help="Show a live plot of the data as it comes from the serial connection")
parser.add_argument("-p", "--port", type=str, default=None, const=None, nargs="?", help="Set the port to read the serial data from")
parser.add_argument("-r", "--runtime", default=5, type=int, help="Set the time in seconds to stream data from the sensor")
parser.add_argument("-s", "--save", default=False, const=True, type=str, nargs="?", help="Save data to the specified file")
args = parser.parse_args()

# Prompt for a port if none is given
if args.port == None:
    print("Known available ports:")
    for port in serial.tools.list_ports.comports():
        print(port.device)

    args.port = input("Enter the port to read data from: ")

while True:
    try:
        test = serial.Serial(port=args.port, baudrate=args.baudrate)
        test.close()
        break
    except:
        args.port = input("Invalid serial port, enter a valid one: ")

# If neither the save nor the show live plot flag are specified, prompt for at least one of them
if args.live_plot == False and args.save == False:
    while True:
        save = input("Would you like to save the sensor data to a file? (y/N): ")
        if save.lower() == "y":
            args.save = input("Enter a csv file to save the data to: ")
            break
        elif save.lower() == "n" or save == "":
            break

        print("Please enter y or n")

    while True:
        live_plot = input("Would you like to show a live plot of the data? (y/N): ")
        if live_plot.lower() == "y":
            args.live_plot = True
            break
        elif live_plot.lower() == "n" or live_plot == "":
            break

        print("Please enter y or n")

    if args.live_plot == False and args.save == False:
        print("Nothing to do, exiting")
        sys.exit(0)

if args.save == True:
    args.save = input("Enter a csv file to save the data to: ")

# Close everything nicely when control-C is pressed
def signal_handler(signal, frame):
    if args.live_plot:
        plt.close()
    if args.save:
        f.close()
    ts.stop = True
    sys.exit(0)
signal.signal(signal.SIGINT, signal_handler)

# Set up live plot
if args.live_plot:
    # The values read from the serial data and the relative times they were recieved
    shear_values = [[], []] 
    force_values = [[], [], [], []] 
    times = []

    # Set up matplotlib graph
    fig = plt.figure()
    shear_plot = fig.add_subplot(211)
    force_plot = fig.add_subplot(212)
    shear_plot.cla()
    force_plot.cla()
    shear_plot.hold(True)
    shear_plot.hold(True)
    shear_plot.set_title("Shear force")
    force_plot.set_title("Force")

    # Label and color the lines
    shear_lines = shear_plot.plot(times, shear_values[0], "r-",
                                  times, shear_values[1], "g-")
    force_lines = force_plot.plot(times, force_values[0], "b-",
                                  times, force_values[1], "c-",
                                  times, force_values[2], "m-",
                                  times, force_values[3], "y-")

    shear_plot.legend(shear_lines, ["Shear X", "Shear Y"], loc="upper left")
    force_plot.legend(force_lines, ["Upper-right Tile (1)", "Lower-right Tile (2)", "Lower-left Tile (3)", "Upper-left Tile (4)"], loc="upper left")

    def _blit_draw(self, artists, bg_cache):
        updated_ax = []
        for a in artists:
            if a.axes not in bg_cache:
                try:
                    a.set_data([], [])
                except:
                    pass
        for a in artists:
            if a.axes not in bg_cache:
                bg_cache[a.axes] = a.figure.canvas.copy_from_bbox(a.axes.figure.bbox)
            a.axes.draw_artist(a)
            updated_ax.append(a.axes)

        for ax in set(updated_ax):
            ax.figure.canvas.blit(ax.figure.bbox)

    matplotlib.animation.Animation._blit_draw = _blit_draw

    if args.disable_axes:
        shear_plot.xaxis.set_visible(False)
        force_plot.xaxis.set_visible(False)
    else:
        shear_plot.xaxis.set_animated(True)
        force_plot.xaxis.set_animated(True)

# Set up file to save to
if args.save:
    f = open(args.save, "w")
    f.write("Time, Shear X, Shear Y, Tile 1, Tile 2, Tile 3, Tile 4\n")

def animate(num, ts):
    if num > 1:
        try:
            # Copy over the data from the sensor and trim them to be the same length
            # This must be done so that the data doesn't get updated on the sensor in
            # the middle of plotting, which causes matplotlib to throw errors because
            # the data and time lists will have different lengths
            ts_data = copy.copy(ts.data)
            ts_times = copy.copy(ts.times)
            while len(ts_data) > len(ts_times):
                ts_data = ts_data[0:-1]

            # Update times
            for i in range(len(ts_times)):
                times.append(ts_times[i])

            # Remove data points and times no longer on the graph
            while len(times) > 1 and times[1] < times[-1] - args.keep:
                times.pop(0)
                for i in range(2):
                    shear_values[i].pop(0)
                for i in range(4):
                    force_values[i].pop(0)

            # Plot shear data
            for i in range(2):
                for j in range(len(ts_data)):
                    shear_values[i].append(ts_data[j][i])

                # Update graph with new data points
                shear_lines[i].set_xdata(times)
                shear_lines[i].set_ydata(shear_values[i])

                # Draw line
                shear_lines[i].set_data(times, shear_values[i])

            # Plot force data
            for i in range(4):
                for j in range(len(ts_data)):
                    force_values[i].append(ts_data[j][i+2])

                # Update graph with new data points
                force_lines[i].set_xdata(times)
                force_lines[i].set_ydata(force_values[i])

                # Draw line
                force_lines[i].set_data(times, force_values[i])

            # Set axis dimensions
            shear_plot.axis([times[-1] - args.keep, times[-1], -500, 500])
            force_plot.axis([times[-1] - args.keep, times[-1], -500, 500])

            # Clear time and data cache from the sensor
            ts.data = []
            ts.times = []
        except Exception as e: # Sometimes fails when corrupted data is recieved
            print(e)
    
    return shear_lines + force_lines + [shear_plot.xaxis, force_plot.xaxis]

def save_data(time, data):
    data_str = str(time) + ", "
    for i in data:
        data_str += str(i) + ", "
    f.write(data_str[:-2] + "\n")

def sensor_callback(time, data):
    if args.save:
        save_data(time, data)

# Set up tactile sensor
ts = sensor.TactileSensor(callback=sensor_callback, port=args.port, runtime=args.runtime)

try:
    ts.Initialize()
    ani = animation.FuncAnimation(fig, animate, fargs=[ts], blit=True, interval=10)

    sensor_thread = threading.Thread(target=ts.run, name="sensor")
    sensor_thread.start()
    plt.show()
except Exception as e:
    print(e)
finally:
    if args.live_plot:
        plt.close()
    if args.save:
        f.close()
    ts.stop = True
